import torch
import torch.nn as nn
from abc import abstractmethod
import os
import inspect

class BaseModel(nn.Module):
    """
    Abstract base class for all models.
    Requires implementation of the calc_loss function.
    Implements weight intialization functions.
    """

    def __init__(self, **args):
        super().__init__()

    def __init_subclass__(cls, **kwargs):
        super().__init_subclass__(**kwargs)
        init_signature = inspect.signature(cls.__init__)
        parameters = list(init_signature.parameters.values())

        # Check if the __init__ method accepts variable keyword arguments (**kwargs)
        if not any(param.kind == inspect.Parameter.VAR_KEYWORD
                   for param in init_signature.parameters.values()):
            error = f"The __init__ method of '{cls.__name__}' must accept **kwargs."+\
                "\nPlease add **kwargs to the __init__ method."+\
                "\nThis is because the Tuner needs to be able to initialize different models."
            raise TypeError(error)

        # Check all parameters after 'self'
        # The first parameter is expected to be 'self'
        for param in parameters[1:]:
            # Disallow positional-only and positional-or-keyword arguments
            if param.kind in (
                inspect.Parameter.POSITIONAL_ONLY,
                inspect.Parameter.POSITIONAL_OR_KEYWORD,
            ):
                raise TypeError(
                    f"The __init__ method of '{cls.__name__}' can only have "
                    f"keyword-only arguments, but '{param.name}' is not."
                )

    @abstractmethod
    def calc_loss(self, batch, criterion, **kwargs) -> torch.Tensor:
        """
        Calculates the loss for a given batch and criterion.

        Parameters
        ----------
        batch: torch.Tensor
            A batch of data.
        criterion: torch.nn.Module
            A loss function.
        **kwargs: dict
            Additional arguments if needed.

        Returns
        -------
        torch.Tensor
            The loss for the batch.
        """
        pass


    def kaiming_init(m):
        """
        Kaiming initialization for linear layers with ReLU
        """
        if type(m) is nn.Linear:
            torch.nn.init.kaiming_uniform_(m.weight, nonlinearity='relu')
            m.bias.data.fill_(0.01)

    def xavier_init(m):
        """
        Xavier initialization for linear layers with ReLU
        """
        if type(m) is nn.Linear:
            torch.nn.init.xavier_normal_(m.weight)
            m.bias.data.fill_(0.01)

    def proper_weight_init(self, sample = None):
        """
        Proper default weight initialization.
        If there are uninitialized lazy layers, initialize them by providing a sample.
        If there are more ReLU-like activations than sigmoid-like activations, initialize the linear layers with Kaiming initialization,
        else Xavier initialization.

        Parameters
        ----------
        sample: torch.Tensor
            A sample of the data.
        """

        has_lazy_layers_before = False
        for module in self.modules():
            if isinstance(module, nn.modules.lazy.LazyModuleMixin) and module.has_uninitialized_params():
                has_lazy_layers_before = True
                break

        if has_lazy_layers_before:
            if sample is None:
                raise ValueError("Sample must be provided if there are uninitialized lazy layers.")

            try:
                with torch.no_grad():
                    self(sample)
            except Exception as e:
                raise ValueError(f"Error during initialization: {e}")


        amount_relu_like_activations = 0
        amount_sigmoid_like_activations = 0
        for module in self.modules():
            if isinstance(module, (nn.ReLU, nn.ReLU6, nn.LeakyReLU, nn.ELU, nn.PReLU, nn.GELU)):
                amount_relu_like_activations += 1
            elif isinstance(module, (nn.Sigmoid, nn.Tanh)):
                amount_sigmoid_like_activations += 1

        for name, param in self.named_parameters():
            if 'weight' in name:
                if 'norm' in name:  # Layer norm weights
                    nn.init.ones_(param)
                elif param.dim() >= 2:  # For matrices (linear layers)
                    if amount_relu_like_activations > amount_sigmoid_like_activations:
                        torch.nn.init.kaiming_uniform_(param, nonlinearity='relu')
                    else:
                        torch.nn.init.xavier_normal_(param)
                else:  # For vectors
                    nn.init.uniform_(param, -0.1, 0.1)
            elif 'bias' in name:
                nn.init.zeros_(param)  # Always initialize biases to zero


    def save_model(self, path):
        """
        Save models state to given path.
        Will create directories if necessary.

        Parameters
        ----------
        path: str
            Path to save model to.
        """

        base_dir = os.path.dirname(path)
        if not os.path.exists(base_dir):
            os.makedirs(base_dir)
        torch.save(self.state_dict(), path)

    def load_model(self, path):
        """
        Load model from given path.

        Parameters
        ----------
        path: str
            Path to load model from.
        """
        self.load_state_dict(torch.load(path))




import torch

class TotalOptimizer:

    """
    Unifies the optimizer, the scheduler, gradient clipping and more.
    """

    def __init__(self, model, optimizer, scheduler=None, max_norm=None):
        self.model = model
        self.optimizer = optimizer
        self.scheduler = scheduler
        self.max_norm = max_norm

    def step(self):
        if self.max_norm:
            torch.nn.utils.clip_grad_norm_(self.model.parameters(), self.max_norm)
        self.optimizer.step()
        if self.scheduler:
            self.scheduler.step()

    def zero_grad(self):
        self.optimizer.zero_grad()


def get_optimizer_total_optimizer(model, learning_rate=1e-3,
                                  optimizer_method="Adam",
                                  step_size=10,
                                  gamma=0.8,
                                  max_norm=1.0,
                                  use_scheduler=False,
                                  use_grad_clip=False,
                                  optimizer_kwargs={},
                                  **kwargs) -> TotalOptimizer:
    """
    Get the optimizer and the scheduler

    Parameters
    ----------
    model: torch.nn.Module
        The model to train.
    learning_rate: float
        The learning rate.
    optimizer_method: str
        The optimizer method.
        Respective default value is "Adam"
    step_size: int
        The step size for the scheduler.
        Respective default value is 10
    gamma: float
        The gamma for the scheduler.
        Respective default value is 0.8
    max_norm: float
        The max norm for the gradient clipping.
        Respective default value is 1.0
    use_scheduler: bool
        Whether to use a scheduler.
        If True, the scheduler will initialized using step_size and gamma.
        Respective default values are 10 and 0.8.
    use_grad_clip: bool
        Whether to use gradient clipping.
        If True, the gradient clipping will initialized using max_norm.
        Respective default value is 1.0
    optimizer_kwargs: dict
        The kwargs for the optimizer.
        Respective default value is {}

    Returns
    -------
    TotalOptimizer
    """

    scheduler = None

    if optimizer_method == "SGD":
        optimizer = torch.optim.SGD(model.parameters(), learning_rate, **optimizer_kwargs)
    else:
        optimizer = torch.optim.Adam(model.parameters(), learning_rate, **optimizer_kwargs)
    if use_scheduler:
        scheduler = torch.optim.lr_scheduler.StepLR(optimizer, step_size=step_size, gamma=gamma)
    if not use_grad_clip:
        max_norm = None

    return TotalOptimizer(model, optimizer, scheduler, max_norm)



from tqdm import tqdm
import torch
import numpy as np
from my_pytorch_kit.train.optimizers import TotalOptimizer
from my_pytorch_kit.model.models import BaseModel
from my_pytorch_kit.train.tensorboard import get_tensorboard_logger

def create_tqdm_bar(iterable, desc):
    return tqdm(enumerate(iterable), total=len(iterable), desc=desc, dynamic_ncols=True)



class Trainer:
    """
    A class for training a model.
    """

    def __init__(self, model, train_loader, val_loader, tb_logger = None, initialize_tb = True):
        """
        Initialize the trainer.

        Parameters
        ----------
        model: torch.nn.Module
            The model to train.
        train_loader: torch.utils.data.DataLoader
            The training data loader.
        val_loader: torch.utils.data.DataLoader
            The validation data loader.
        tb_logger: SummaryWriter
            The tensorboard logger.
        initialize_tb: bool
            Whether to initialize the tensorboard logger, if tb_logger is None.
        """
        self.model = model
        self.train_loader = train_loader
        self.val_loader = val_loader

        if tb_logger is None and initialize_tb:
            tb_logger = get_tensorboard_logger()
        self.tb_logger = tb_logger

        self.isTotalOptimizer = True



    def train(self, 
              optimizer,
              loss_func,
              epochs=10,
              loss_cutoff_rate=0.1,
              patience=None,
              name="model",
              override_instance_errors=False, **kwargs) -> float:
        """
        Train a model and log loss to tensorboard.
        If interrupted by KeyboardInterrupt, exit gracefully.
        Also compatible with torch.nn.Model as a model and torch.optim.Optimizer as an optimizer,
        if override_instance_errors is True.
        Returns best validation loss.

        Parameters
        ----------
        optimizer: torch.optim.Optimizer
            The optimizer.
        loss_func: torch.nn.Module
            The loss function.
        epochs: int
            The number of epochs to train for. (default: 10)
        loss_cutoff_rate: float
            The rate at which to cut off the loss. (default: 0.1)
        patience: int
            The number of epochs to wait before early stopping. (default: None)
        name: str
            The name of the model.
        override_instance_errors: bool
            Whether to ignore instance errors.

        Returns
        -------
        best_val_loss: float
            The best validation loss.
        """

        if not issubclass(type(self.model), BaseModel) and not override_instance_errors:
            raise ValueError(f"model with type {type(self.model)} must be a subclass of BaseModel")

        if not isinstance(optimizer, TotalOptimizer):
            if not override_instance_errors:
                raise ValueError("Optimizer must be an instance of TotalOptimizer")
            self.isTotalOptimizer = False

        loss_cutoff = int(len(self.train_loader) * loss_cutoff_rate)
        patience_counter = 0
        best_val_loss = float('inf')

        try:
            for epoch in range(epochs):
                training_loss = []
                validation_loss = []

                # TRAINING
                self.model.train()
                training_loop = create_tqdm_bar(self.train_loader, desc=f'Training Epoch [{epoch + 1}/{epochs}]')
                for train_iteration, batch in training_loop:
                    optimizer.zero_grad()
                    loss = self.model.calc_loss(batch, loss_func)
                    loss.backward()
                    optimizer.step()


                    training_loss.append(loss.item())
                    last_cutoff_loss = np.mean(training_loss[-loss_cutoff:])

                    # Update the progress bar.
                    self.set_training_post_fix(training_loop, last_cutoff_loss, optimizer)

                    # Update the tensorboard logger.
                    self.add_tb_scalar(f'{name}/train_loss', loss.item(), epoch * len(self.train_loader) + train_iteration)

                # VALIDATION
                self.model.eval()
                val_loop = create_tqdm_bar(self.val_loader, desc=f'Validation Epoch [{epoch + 1}/{epochs}]')

                with torch.no_grad():
                    for val_iteration, batch in val_loop:
                        loss = self.model.calc_loss(batch, loss_func)
                        validation_loss.append(loss.item())
                        last_cutoff_loss = np.mean(validation_loss[-loss_cutoff:])

                        # Update the progress bar.
                        self.set_validation_post_fix(val_loop, last_cutoff_loss)

                        # Update the tensorboard logger.
                        self.add_tb_scalar(f'{name}/val_loss', last_cutoff_loss, epoch * len(self.val_loader) + val_iteration)

                # best val loss check
                total_val_loss = np.mean(validation_loss)
                if total_val_loss < best_val_loss:
                    best_val_loss = total_val_loss
                    patience_counter = 0
                else:
                    patience_counter += 1

                # patience check
                if patience is not None and patience_counter >= patience:
                    print(f"Early stopping at epoch {epoch + 1}.")
                    break


        except KeyboardInterrupt:
            print("Training interrupted by user.")

        return best_val_loss



    def add_tb_scalar(self, *args, **kwargs):
        if self.tb_logger is not None:
            self.tb_logger.add_scalar(*args, **kwargs)

    def set_training_post_fix(self, loop, loss, optimizer):
        if self.isTotalOptimizer:
            lr = optimizer.optimizer.param_groups[0]['lr']
        else:
            lr = optimizer.param_groups[0]['lr']
        loop.set_postfix(loss = "{:.4f}".format(loss),
                         lr = "{:.4f}".format(lr))

    def set_validation_post_fix(self, loop, loss):
        loop.set_postfix(loss = "{:.4f}".format(loss))




from torch.utils.tensorboard import SummaryWriter
import os

def get_tensorboard_logger(path="logs") -> SummaryWriter:
    """
    Get a tensorboard logger

    Parameters
    ----------
    path: str
        The path to the logs directory

    Returns
    -------
    SummaryWriter
    """
    num_of_runs = len(os.listdir(path)) if os.path.exists(path) else 0
    path = os.path.join(path, f'run_{num_of_runs + 1}')
    return SummaryWriter(log_dir=path)
